from base_st_classes import *

class SetattrSuite(NFSSuite):
    """Test operation 34: SETATTR

    FIXME: Test invalid filehandle. 

    Equivalence partitioning:

    Input Condition: current filehandle
        Valid equivalence classes:
            file(10)
            dir(11)
            block(12)
            char(13)
            link(14)
            socket(15)
            FIFO(16)
            attribute directory(17)
            named attribute(18)
        Invalid equivalence classes:
            no filehandle(19)
    Input Condition: stateid
        Valid equivalence classes:
            all bits zero(20)
            all bits one(21)
            valid stateid from open(22)
        Invalid equivalence classes:
            invalid stateid(23)
    Input Condition: obj_attributes.attrmask
        Valid equivalence classes:
            writeable attributes without object_size(30)
            writeable attributes with object_size(31)
        Invalid equivalence classes:
            non-writeable attributes(32)
    Input Condition: obj_attributes.attr_vals
        Valid equivalence classes:
            valid attributes(40)
        Invalid equivalence classes:
            invalid attributes(41)
            attr with invalid utf8(42)
    """

    def setUp(self):
        NFSSuite.setUp(self)
        self.fh = None
        self.stateid = None

    def tearDown(self):
        self.ncl.close_file(self.fh, self.stateid, error=None)
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    def _setattr(self, stateid=None, file=None, attrdict={FATTR4_MODE:0775},
                 error=[NFS4_OK]):
        if stateid==None: stateid = stateid4(0, "")
        if file==None: file = self.tmp_dir + [self.obj_name]
        
        lookupops = self.ncl.lookup_path(file)
        ops = [self.ncl.putrootfh_op()] + lookupops

        attrs = nfs4lib.dict2fattr(attrdict)
        ops.append(self.ncl.setattr_op(stateid, attrs))
        oldres = self.ncl.do_ops(ops)
        #if self._check_notsupp(res): return

        self.assert_status(oldres, error)
        # If succeeded, check that getattr returns new values
        if oldres.status==NFS4_OK and FATTR4_TIME_MODIFY_SET not in attrdict.keys():
            ops = [self.ncl.putrootfh_op()] + lookupops
            ops.append(self.ncl.getattr(attrdict.keys()))
            res = self.ncl.do_ops(ops)
            self.assert_OK(res)
            obj = res.resarray[-1].arm.arm.obj_attributes
            d = nfs4lib.fattr2dict(obj)
            for key in d.keys():
                setval = attrdict[key]
                if d[key] != setval:
                    self.fail("GETATTR after SETATTR was not consistent, SET(%s)=%d, GET(%s)=%d"%(key,setval,key,d[key]))
        return oldres
    
    #
    # Testcases covering valid equivalence classes.
    #
    def testStateidOnes(self):
        """SETATTR(FATTR4_MODE) on regular file with stateid=ones
        
        Covered valid equivalence classes: 10, 21, 30, 40
        """
        # FRED FIXME - this is useless, since stateid only relevant with size
        
        self.init_connection()
        # Create a file with mode=0644
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Set mode to 0755
        stateval = nfs4lib.long2opaque(0xffffffffffffffffffffffffL)
        stateid =  stateid4(0xffffffffL, stateval)
        self._setattr(stateid)
        
    def testDir(self):
        """SETATTR(FATTR4_MODE) on directory

        Covered valid equivalence classes: 11, 20, 30, 40
        """
        try: self.create_obj()
        except SkipException, e:
            self.skip(e)

        self._setattr()

    def testBlock(self):
        """SETATTR(FATTR4_MODE) on block device

        Covered valid equivalence classes: 12, 20, 30, 40
        """
        try: self.create_obj(type = NF4BLK)
        except SkipException, e:
            self.skip(e)

        self._setattr()

    def testChar(self):
        """SETATTR(FATTR4_MODE) on char device

        Covered valid equivalence classes: 13, 20, 30, 40
        """
        try: self.create_obj(type = NF4CHR)
        except SkipException, e:
            self.skip(e)

        self._setattr()

    def testLink(self):
        """SETATTR(FATTR4_MODE) on symbolic link

        Covered valid equivalence classes: 14, 20, 30, 40

        Comments: The response to mode setting on a symbolic link is
        server dependent; about any response is valid. We just test
        and print the result. 
        """
        try: self.create_obj(type = NF4LNK)
        except SkipException, e:
            self.skip(e)

        res = self._setattr()

        self.info_message("SETATTR(FATTR4_MODE) on symlink returned %s" \
                          % nfsstat4_id[res.status])
        
    def testSocket(self):
        """SETATTR(FATTR4_MODE) on socket

        Covered valid equivalence classes: 15, 20, 30, 40
        """
        try: self.create_obj(type = NF4SOCK)
        except SkipException, e:
            self.skip(e)

        res = self._setattr()

    def testFIFO(self):
        """SETATTR(FATTR4_MODE) on FIFO

        Covered valid equivalence classes: 16, 20, 30, 40
        """
        try: self.create_obj(type = NF4FIFO)
        except SkipException, e:
            self.skip(e)

        res = self._setattr()

    def xtestNamedattrdir(self):
        """SETATTR(FATTR4_MODE) on named attribute directory

        Covered valid equivalence classes: 17, 20, 30, 40
        """
        # FIXME: Implement.
        self.skip("(TEST NOT IMPLEMENTED)")

    def xtestNamedattr(self):
        """SETATTR(FATTR4_MODE) on named attribute 

        Covered valid equivalence classes: 18, 20, 30, 40
        """
        # FIXME: Implement.
        self.skip("(TEST NOT IMPLEMENTED)")

    def testChangeSize(self):
        """SETATTR(FATTR4_SIZE) with changes to file size and valid stateid

        Covered valid equivalence classes: 10, 22, 31, 40
        """
        # Create a file with mode=0644
        self.init_connection()
        fh,stateid = self.ncl.create_confirm()

        # Verify size is 0
        lookupops = self.ncl.lookup_path(self.tmp_dir + [self.obj_name])
        ops = [self.ncl.putrootfh_op()] + lookupops
        ops.append(self.ncl.getattr([FATTR4_SIZE]))
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        obj = res.resarray[-1].arm.arm.obj_attributes
        d = nfs4lib.fattr2dict(obj)
        self.assertEqual(d[FATTR4_SIZE],0,"Size does not start at 0")

        # Set new size
        self._setattr(stateid,attrdict={FATTR4_SIZE:1000})

    def testChangeObjSize(self):
        """SETATTR(FATTR4_SIZE) on a nonfile objects should fail"""
        stateid = stateid4(0, "")
        for type in [NF4DIR,NF4SOCK,NF4FIFO,NF4LNK,NF4BLK,NF4CHR]:
            try: self.create_obj()
            except SkipException, e:
                self.skip(e)
            self._setattr(stateid, attrdict={FATTR4_SIZE:1000},
                          error=[NFS4ERR_BAD_STATEID])
            self.remove_object()

    #
    # Testcases covering invalid equivalence classes.
    #
    def testNoFh(self):
        """SETATTR without (cfh) should return NFS4ERR_NOFILEHANDLE

        Covered invalid equivalence classes: 19
        """
        stateid = stateid4(0, "")
        attrs = nfs4lib.dict2fattr({FATTR4_MODE:0755})
        ops = [self.ncl.setattr_op(stateid, attrs)]
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NOFILEHANDLE])

    def testBadStateid1(self):
        """SETATTR with bad stateid w/o setting size should return NFS4_OK

        Covered invalid equivalence classes: 23
        
        NOTE Mangling functions look into server opaque data, thus are
        specific to the CITI linux server

        Actually, the RFC states 'When the file size attribute is not set,
        the special stateid consisting of all bits zero should be passed.',
        so a non-OK response would be valid.
        """
        # Get a valid stateid
        self.init_connection()
        self.fh, self.stateid = self.ncl.create_confirm()

        # Mangle it
        badid = self.makeBadID(self.stateid)

        self._setattr(badid)

    def testBadStateid2(self):
        """SETATTR with bad stateid and setting size should return NFS4ERR_BAD_STATEID

        Covered invalid equivalence classes: 23
        
        NOTE Mangling functions look into server opaque data, thus are
        specific to the CITI linux server
        """
        # Get a valid stateid
        self.init_connection()
        self.fh, self.stateid = self.ncl.open_confirm(self.regfile[:-1],
                                                      self.regfile[-1])
        # Mangle it
        badid = self.makeBadID(self.stateid)

        self._setattr(badid, self.regfile, {FATTR4_SIZE:1000},
                      error=[NFS4ERR_BAD_STATEID])

    def testNonWriteable(self):
        """SETATTR on a read-only attr should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 32
        """
        self.init_connection()
        # Create a file with mode=0644
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)


        lookupops = self.ncl.lookup_path(self.regfile)

        for attrdict in [ {FATTR4_LINK_SUPPORT:TRUE},
                          {FATTR4_SYMLINK_SUPPORT:FALSE},
                          {FATTR4_NAMED_ATTR:FALSE},
                          {FATTR4_UNIQUE_HANDLES:FALSE},
                          {FATTR4_CANSETTIME:FALSE},
                          {FATTR4_CASE_INSENSITIVE:FALSE},
                          {FATTR4_CASE_PRESERVING:FALSE},
                          {FATTR4_CHOWN_RESTRICTED:FALSE} ]:
            self._setattr(attrdict=attrdict,error=[NFS4ERR_INVAL])

    def testInvalidAttr(self):
        """SETATTR with invalid attribute data should return NFS4ERR_BADXDR

        Covered invalid equivalence classes: 41

        Comments: This testcase try to set FATTR4_MODE but does not send any
        mode data. The server should return NFS4ERR_BADXDR. 
        """
        lookupops = self.ncl.lookup_path(self.regfile)
        ops = [self.ncl.putrootfh_op()] + lookupops

        stateid = stateid4(0, "")
        attrmask = nfs4lib.list2attrmask([FATTR4_MODE])
        attr_vals = ""
        obj_attrs = fattr4(attrmask, attr_vals)
        ops.append(self.ncl.setattr_op(stateid, obj_attrs))

        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_BADXDR])

    def xtestInvalidAttr2(self):
        """SETATTR with invalid attribute data should return NFS4ERR_BADXDR

        Covered invalid equivalence classes: 41

        Comments: This testcase try to set FATTR4_MODE but does not send any
        mode data. The server should return NFS4ERR_BADXDR. 
        """
        # Create a file with mode=0644
        self.init_connection()
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        self._setattr(attrdict={FATTR4_MODE:0777777},error=[NFS4_OK])

    def testNonUTF8(self):
        """SETATTR(FATTR4_MIMETYPE) with non-utf8 string should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 42

        """
        for name in self.get_invalid_utf8strings():
            self._setattr(file=self.regfile,attrdict={FATTR4_MIMETYPE:name},
                          error=[NFS4_OK,NFS4ERR_ATTRNOTSUPP])
    #
    # Extra tests. 
    #
    def testInvalidTime(self):
        """SETATTR(FATTR4_TIME_MODIFY_SET) with invalid nseconds

        Extra test

        nseconds larger than 999999999 are considered invalid.
        SETATTR(FATTR4_TIME_MODIFY_SET) should return NFS4ERR_INVAL on
        such values. 
        """
        # Create a file with mode=0644
        self.init_connection()
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # First, try to set the date to 500 000 000 = 1998-07-09
        # to check if setting time_modify is possible at all. 
        time = nfstime4(seconds=500000000, nseconds=0)
        settime = settime4(set_it=SET_TO_CLIENT_TIME4, time=time)
        res = self._setattr(attrdict={FATTR4_TIME_MODIFY_SET:settime},
                            error=[NFS4_OK,NFS4ERR_ATTRNOTSUPP])
        if res.status == NFS4ERR_ATTRNOTSUPP:
            self.info_message("Attribute time_modify_set is not supported, "
                              "skipping test")
            return

        # Ok, lets try nseconds = 1 000 000 000
        time = nfstime4(seconds=500000000, nseconds=int(1E9))
        settime = settime4(set_it=SET_TO_CLIENT_TIME4, time=time)
        res = self._setattr(attrdict={FATTR4_TIME_MODIFY_SET:settime},
                            error=[NFS4ERR_INVAL])

    def testUnsupported(self):
        """SETATTR with unsupported attr should return NFS4ERR_ATTRNOTSUPP
        """
        
        try: self.create_obj()
        except SkipException, e:
            self.skip(e)

        # Get list of supported attributes
        supported = self.supportedAttrs()
        
        # Try to set any of the writable attrs that aren't supported
        attrbitnum_dict = nfs4lib.get_attrbitnum_dict()
        flag = 0
        for (attr,value) in self.ncl.writable_attrs:
            bitnum = attrbitnum_dict[attr]
            if not (supported & (2**bitnum)):
                flag = 1
                self._setattr(attrdict={bitnum:value},error=[NFS4ERR_ATTRNOTSUPP])
        if not flag:
            self.skip("No unsupported writable attributes")

    def testInodeLocking(self):
        """SETATTR: This causes printk message due to inode locking bug

        nfsd: inode locked twice during operation.
        Sporadic system crashes can occur after running this test
        """
        self.fail("Test skipped, inode locking bug causes sporadic system crashes")
        self.init_connection()
        try: self.create_obj("dir")
        except SkipException, e:
            self.skip(e)

        newdir = self.tmp_dir + ["dir"]
        self.fh, self.stateid = self.ncl.create_confirm(dir=newdir)

        # In a single compund statement, setattr on dir and then
        # do a state operation on a file in dir
        stateid = stateid4(0, "")
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(newdir)
        obj_attributes = nfs4lib.dict2fattr({FATTR4_MODE: 0755})
        ops.append(self.ncl.setattr_op(stateid, obj_attributes))
        ops.append(self.ncl.lookup_op(self.obj_name))
        ops.append(self.ncl.write("blah",self.stateid)) #remove will also err
        res = self.ncl.compound(ops)
        self.assert_OK(res)
        self.fail("This causes printk message")

