from base_st_classes import *

class SetclientidSuite(NFSSuite):
    """Test operation 35: SETCLIENTID

    FIXME: Test cases that trigger NFS4ERR_CLID_INUSE. 

    Equivalence partitioning:

    Input Condition: client.verifier
        Valid equivalence classes:
            all input(10)
        Invalid equivalence classes:
            -
    Input Condition: client.id
        Valid equivalence classes:
            all input(20)
        Invalid equivalence classes:
            -
    Input Condition: callback.cb_program
            all input(30)
        Invalid equivalence classes:
            -
    Input Condition: callback.cb_location
            all input(40)
        Invalid equivalence classes:
            -

    Comments: If client has never connected to the server, every
    client.verifier and client.id is valid. All callback data is also
    allowed as input, but failing to provide the correct address means
    callbacks will not be used. 
    """
    #
    # Testcases covering valid equivalence classes.
    #
    def testValid(self):
        """SETCLIENTID simple call

        Covered valid equivalence classes: 10, 20, 30, 40
        """
        # client
        verifier = self.ncl.gen_random_64()
        id = self.ncl.gen_uniq_id()
        client = nfs_client_id4(verifier, id)

        # callback
        cb_program = 0
        r_netid = "udp"
        # FIXME
        # Internet, Port number, IP, RFU
        r_addr = "0002" + "0000" + "00112233" + "00" 
        cb_location = clientaddr4(r_netid, r_addr)
        callback = cb_client4(cb_program, cb_location)
        
        setclientidop = self.ncl.setclientid_op(client, callback)
        res = self.ncl.do_ops([setclientidop])
        self.assert_OK(res)

    #
    # Extra tests.
    #
    def _set(self, ncl, id, verifier):
        setclientidop = ncl.setclientid(verifier,id)
        res = ncl.do_ops([setclientidop])
        return res

    def _confirm(self, ncl, clientid, verifier):
        setclientid_confirmop = ncl.setclientid_confirm_op(clientid,verifier)
        res = ncl.do_ops([setclientid_confirmop])
        self.assert_OK(res)
    
    def testInUse(self):
        """SETCLIENTID with same nfs_client_id.id should return NFS4ERR_CLID_INUSE
        This requires NCL1 and NCL2 to have different principals (UIDs).

        Extra test
        """

        if self.ncl.sec[0] != 1:
            self.skip("Can't use this test with security flavor %i"
                      % self.ncl.sec[0])

        # If use same id as other tests, and 2nd try works, bad things happen
        id = self.ncl.gen_uniq_id() + "setclientid.testInUse"
        verifier = self.ncl.gen_random_64()
        
        # 1st SETCLIENTID + SETCLIENTID_CONFIRM
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)
        clientid = res.resarray[0].arm.arm.clientid
        idconfirm = res.resarray[0].arm.arm.setclientid_confirm
        self._confirm(self.ncl, clientid, idconfirm)

        # 2nd SETCLIENTID 
        res = self._set(self.ncl2, id, verifier)
        self.assert_status(res, [NFS4ERR_CLID_INUSE])

    def testLoseAnswer(self):
        """SETCLIENTID after a client reboot could cause case not covered in RFC
        """

        id = self.ncl.gen_uniq_id() + "setclientid.testLoseAnswer"
        verifier = self.ncl.gen_random_64()

        # 1st SETCLIENTID + SETCLIENTID_CONFIRM
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)
        clientid = res.resarray[0].arm.arm.clientid
        idconfirm = res.resarray[0].arm.arm.setclientid_confirm
        self._confirm(self.ncl, clientid, idconfirm)

        # 2nd SETCLIENTID 
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)

        # Now assume client reboot, id should stay same, but verifier changes,
        # and we have lost result from second setclientid.
        # This case is not covered in RFC 3530.
        verifier = self.ncl.gen_random_64()
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)

    def testAllCases(self):
        """SETCLIENTID with each case from RFC"""

        id = self.ncl.gen_uniq_id() + "setclientid.testAllCases"
        verifier = self.ncl.gen_random_64()

        # no (*x***),no (*x****)
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)
        # no (*x***), (*x****)
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)

        clientid = res.resarray[0].arm.arm.clientid
        idconfirm = res.resarray[0].arm.arm.setclientid_confirm
        self._confirm(self.ncl, clientid, idconfirm)

        # (vxc**), no (vxc**)
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)
        # (vxc**), (vxc**)
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)

        clientid = res.resarray[0].arm.arm.clientid
        idconfirm = res.resarray[0].arm.arm.setclientid_confirm
        self._confirm(self.ncl, clientid, idconfirm)

        verifier = self.ncl.gen_random_64()
        # (*x***), no (*x***)
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)
        # (*xc*s), (*xd*t)
        res = self._set(self.ncl, id, verifier)
        self.assert_OK(res)


    def testLotsOfClients(self):
        """SETCLIENTID called multiple times"""
        id =  self.ncl.gen_uniq_id()
        for i in range(1024):
#           id = "adfg"+str(i)+"fgjg"
            verifier = self.ncl.gen_random_64()
            res = self._set(self.ncl, id, verifier)
            self.assert_OK(res)
